#define F_CPU   12000000L  

#include <avr/io.h>
#include <avr/delay.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "gamepad.h"
#include "leds.h"
#include "snes.h"

#define REPORT_SIZE		3
#define GAMEPAD_BYTES	2

/******** IO port definitions **************/
#define SNES_LATCH_DDR	DDRC
#define SNES_LATCH_PORT	PORTC
#define SNES_LATCH_BIT	(1<<4)

#define SNES_CLOCK_DDR	DDRC
#define SNES_CLOCK_PORT	PORTC
#define SNES_CLOCK_BIT	(1<<5)

#define SNES_DATA_PORT	PORTC
#define SNES_DATA_DDR	DDRC
#define SNES_DATA_PIN	PINC
#define SNES_DATA_BIT	(1<<3)

/********* IO port manipulation macros **********/
#define SNES_LATCH_LOW()	do { SNES_LATCH_PORT &= ~(SNES_LATCH_BIT); } while(0)
#define SNES_LATCH_HIGH()	do { SNES_LATCH_PORT |= SNES_LATCH_BIT; } while(0)
#define SNES_CLOCK_LOW()	do { SNES_CLOCK_PORT &= ~(SNES_CLOCK_BIT); } while(0)
#define SNES_CLOCK_HIGH()	do { SNES_CLOCK_PORT |= SNES_CLOCK_BIT; } while(0)

#define SNES_GET_DATA()	(SNES_DATA_PIN & SNES_DATA_BIT)

/*********** prototypes *************/
static void snesInit(void);
static void snesUpdate(void);
static char snesChanged(void);
static void snesBuildReport(unsigned char *reportBuffer);


// the most recent bytes we fetched from the controller
static unsigned char last_read_controller_bytes[GAMEPAD_BYTES];
// the most recently reported bytes
static unsigned char last_reported_controller_bytes[GAMEPAD_BYTES];

static void snesInit(void)
{
	// clock and latch as output
	SNES_LATCH_DDR |= SNES_LATCH_BIT;
	SNES_CLOCK_DDR |= SNES_CLOCK_BIT;
	
	// data as input
	SNES_DATA_DDR &= ~(SNES_DATA_BIT);
	// enable pullup. This should prevent random toggling of pins
	// when no controller is connected.
	SNES_DATA_PORT |= SNES_DATA_BIT;

	// clock is normally high
	SNES_CLOCK_PORT |= SNES_CLOCK_BIT;

	// LATCH is Active HIGH
	SNES_LATCH_PORT &= ~(SNES_LATCH_PORT);

	snesUpdate();
}


/*
 *
       Clock Cycle     Button Reported
        ===========     ===============
        1               B
        2               Y
        3               Select
        4               Start
        5               Up on joypad
        6               Down on joypad
        7               Left on joypad
        8               Right on joypad
        9               A
        10              X
        11              L
        12              R
        13              none (always high)
        14              none (always high)
        15              none (always high)
        16              none (always high)
 *
 */

static void snesUpdate(void)
{
	int i;
	unsigned char tmp=0;

	SNES_LATCH_HIGH();
	_delay_us(12);
	SNES_LATCH_LOW();

	for (i=0; i<8; i++)
	{
		_delay_us(6);
		SNES_CLOCK_LOW();
		
		tmp <<= 1;	
		if (!SNES_GET_DATA()) { tmp |= 1; }

		_delay_us(6);
		
		SNES_CLOCK_HIGH();
	}
	last_read_controller_bytes[0] = tmp;
	for (i=0; i<8; i++)
	{
		_delay_us(6);

		SNES_CLOCK_LOW();

		// notice that this is different from above. We
		// want the bits to be in reverse-order
		tmp >>= 1;	
		if (!SNES_GET_DATA()) { tmp |= 0x80; }
		
		_delay_us(6);
		SNES_CLOCK_HIGH();
	}
	last_read_controller_bytes[1] = tmp;

}

static char snesChanged(void)
{
	static int first = 1;
	if (first) { first = 0;  return 1; }
	
	return memcmp(last_read_controller_bytes, 
					last_reported_controller_bytes, GAMEPAD_BYTES);
}

static void snesBuildReport(unsigned char *reportBuffer)
{
	int x, y;
	unsigned char lrcb1, lrcb2;
	
	if (reportBuffer != NULL)
	{
		lrcb1 = last_read_controller_bytes[0];
		lrcb2 = last_read_controller_bytes[1];

		y = x = 128;
		if (lrcb1&0x1) { x = 255; }
		if (lrcb1&0x2) { x = 0; }
		
		if (lrcb1&0x4) { y = 255; }
		if (lrcb1&0x8) { y = 0; }

		reportBuffer[0]=x;
		reportBuffer[1]=y;

		reportBuffer[2] =	(lrcb1&0x80)>>7;
		reportBuffer[2] |=	(lrcb1&0x40)>>5;
		reportBuffer[2] |=	(lrcb1&0x20)>>3;
		reportBuffer[2] |=	(lrcb1&0x10)>>1;
		reportBuffer[2] |=	(lrcb2&0x0f)<<4;
	}
	memcpy(last_reported_controller_bytes, 
			last_read_controller_bytes, 
			GAMEPAD_BYTES);	
}

const char snes_usbHidReportDescriptor[] PROGMEM = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x05,                    // USAGE (Game Pad)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x09, 0x01,                    //   USAGE (Pointer)
    0xa1, 0x00,                    //   COLLECTION (Physical)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //     LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x02,                    //   REPORT_COUNT (2)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0xc0,                          // END_COLLECTION
    0x05, 0x09,                    // USAGE_PAGE (Button)
    0x19, 0x01,                    //   USAGE_MINIMUM (Button 1)
    0x29, 0x08,                    //   USAGE_MAXIMUM (Button 8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    // REPORT_SIZE (1)
    0x95, 0x08,                    // REPORT_COUNT (8)
    0x81, 0x02,                    // INPUT (Data,Var,Abs)
    0xc0                           // END_COLLECTION
};

Gamepad SnesGamepad = {
	report_size: 		REPORT_SIZE,
	reportDescriptorSize:	sizeof(snes_usbHidReportDescriptor),
	init: 			snesInit,
	update: 		snesUpdate,
	changed:		snesChanged,
	buildReport:		snesBuildReport
};

Gamepad *snesGetGamepad(void)
{
	SnesGamepad.reportDescriptor = (void*)snes_usbHidReportDescriptor;

	return &SnesGamepad;
}

